/*
 * @(#)Maple.java  May 2001
 *
 * This class represents a Maple process.
 */

import java.io.*;
import java.util.*;

public class Maple {

  public Date creationTime;

  public Date lastUseTime;

  public boolean busy = false;

  public boolean condemned = false;

  // An identifier for this process, assigned by the MaplePool
  public int id = 0;

// KM
  // The OS PID for this process, obtained by querying the Maple process
  // a negative value means no process id is available
  public int pid = -1;

  // True if this maple object has been discarded from the Maple pool
  public boolean discarded = false;

  // Since I'm not sure how to throw an exception from one thread to another
  // The MapleCommandThread stores its exceptions in here
  // It is null if no exception was encountered
  public Exception exception = null;
// END KM

  // The number of users assigned to this process
  public int load = 0;

  // The name of the user currently using this process, if any
  public String user;

  private Process process;

  // Output from Java is input to Maple
  private OutputStream mapleInput; 

  // Output from Maple is input to Java
  private BufferedReader mapleOutput;

// KM
  // I think we should handle the Maple Error stream also...
  private BufferedReader mapleError;
// END KM

  // Note: the exec method is passed an argument webOutput.
  // The method reads from mapleOutput, and writes to webOutput.

  // The command used to start Maple.  This should ensure that it starts 
  // in character mode (not using the graphical user interface).
  // The option -q (to suppress the startup message) is added
  // automatically.
  public String program = "maple";

  // Initialization options
  private String initOptions =
     "kernelopts(printbytes=false):" +
     "interface(quiet=true, errorcursor=false, errorbreak=0):";

// KM
  // The command used to kill processes in this OS. 
  // Useful for killing hung maple processes whose PID is known
  public String killProgram = "Process";

  // The argument to the command used to kill processes in this OS.
  // For example "process -k 1234" has killProgram="process" and 
  // killProgramArg="-k" and pid=1234.  In "kill -9 1234" killProgram="kill"
  // killProgramArg="-9" and pid= 1234.
  public String killProgramArg = "-k";
// END KM

  // The exec method asks Maple to executes a command, and then to
  // print the terminator string given below.  The exec method then
  // looks for the terminator string in Maple's output, so see when
  // Maple has finished.
  private static final String 
    terminator = "JAVA: END OF A MAPLE COMMAND IS HERE";

  public Maple() throws IOException {
    this.creationTime = new Date();
    this.lastUseTime = this.creationTime;
    this.startProcess();
  }

  public Maple(String prog) throws IOException {
    this.creationTime = new Date();
    this.lastUseTime = this.creationTime;
    if (prog != null) 
      this.program = prog;
    this.startProcess();
  }

  public Maple(String prog,String initoptions) throws IOException {
    this.creationTime = new Date();
    this.lastUseTime = this.creationTime;
    if (prog != null)
      this.program = prog;
    if (initoptions != null) 
      this.initOptions = initoptions;
    this.startProcess();
  }

// KM
  public Maple(String prog,String initoptions,String killprog,String killprogarg) throws IOException {
    this.creationTime = new Date();
    this.lastUseTime = this.creationTime;
    if (prog != null)
      this.program = prog;
    if (initoptions != null) 
      this.initOptions = initoptions;
    if (killprog != null) 
      this.killProgram = killprog;
    if (killprog != null) 
      this.killProgramArg = killprogarg;
    this.startProcess();
  }
// END KM

  private void startProcess() throws IOException {
    this.lastUseTime = new Date();
    this.process = Runtime.getRuntime().exec(program + " -q");
    this.mapleInput = process.getOutputStream(); 
    this.mapleOutput =
      new BufferedReader(new InputStreamReader(process.getInputStream()));
// KM
    // we may not need this, but it might be a good idea to manually close 
    // it when cleaning up destroyed processes
    this.mapleError =
      new BufferedReader(new InputStreamReader(process.getErrorStream()));
    // get the PID
    this.getPID();
// END KM 
    this.write(initOptions);
  }

// KM
  private void getPID() throws IOException {
    String ms = "if assigned(Sockets) and " +
                "  member(GetProcessID,{exports(Sockets)}) then " +
                "  printf(\"%d\\n\",Sockets:-GetProcessID()) " +
                "else " +
                "  printf(\"-1\\n\") " + 
                "fi;\n";
    this.write(ms);  
    String s = mapleOutput.readLine();
    // if this is not Maple 8 it should not return an integer
    try {
       this.pid = Integer.parseInt(s);
    } catch (Exception e) { 
       this.pid = -1;
    } 
  }
// END KM

  // send a string to Maple
  public void write(String s) throws IOException {
    mapleInput.write(s.getBytes());
    mapleInput.flush();
  }
  
  // return true if the process has exited, or was never started
  // TODO: check what happens if the process is destroyed externally?
  // Should this be synchronized?
  public boolean noProcess() {
    if (process == null) return true;

    try {
      process.exitValue();
    } catch (IllegalThreadStateException ex) {
      return false; // process has not yet exited
    } 

    return true;
  }

  public void checkProcess() throws IOException {
    if (noProcess()) {
      startProcess();
    }
  }

  // Close the Maple session:
  // TODO: check the noProcess behavior when the process is destroyed
  // externally, e.g. from the monitor page.
  public void close() {
    if (noProcess()) {
      // do nothing
    } else {
      try { 
        this.write("shutdown();quit;\n");
        process.waitFor();
      } catch(Exception ex) {}
    }
  }

  public void destroy() {
// KM
    if (pid<0) {
// END KM
      try {
      process.destroy();
// KM
      mapleInput.close();
      mapleOutput.close();
      mapleError.close();
      process = null;
// END KM
      } catch(Exception e) {}
// KM
    } else {
      try {
        // we know the PID so we kill, close, and smash 
        // everything that might possibly cause a problem
        // TODO: What happens if the process is already missing?
        Process rt = Runtime.getRuntime().exec(killProgram + " " +
                                               killProgramArg + " " + 
                                               pid);
        rt.waitFor();
        rt.getOutputStream().close();
        rt.getInputStream().close();
        rt.getErrorStream().close();
        process.destroy();
        mapleInput.close();
        mapleOutput.close();
        mapleError.close();
        process = null;
      } catch (Exception e) { }
    }
// END KM
  }

  // Execute a given command 'cmd' and discard the results
  // The command need not have a semicolon or newline at the end
  public synchronized void exec(String cmd) throws Exception {
    exec(cmd, null);
  }

  // Execute a given command 'cmd' and write the results to 'out'.
  // The command need not have a semicolon or newline at the end
  public synchronized void exec(String cmd, PrintWriter out)
    throws Exception {
  
     lastUseTime = new Date();

    StringBuffer outstring = new StringBuffer();

    // execute user's command:
    this.write(cmd + ";\n");

    // Tell Maple to print the terminator string, so we can see
    // that it has finished.
    this.write( ("printf(\"\\n" + terminator + "\\n\"):\n"));  

    // Next, read the output, line-by-line, until terminator is read:
    while(true) {
        String s = mapleOutput.readLine();

        if (s == null || s.indexOf(terminator) != -1 ) {
          break;
        }

        outstring.append(s);
        outstring.append("\n");
      } 

    // TODO: check if browser accepts <latex> and if so, omit TtH.
    if (out != null) {
      out.print(tth.TtH.convert(outstring.toString()));
    }
  }

  public boolean check() {
    StringWriter s = new StringWriter();
    PrintWriter p = new PrintWriter(s);

// KM
// I don't think we want to count this as a "last use", otherwise refreshing
// the AIM Monitor page resets all the "last use" times to the present time
// I think its more useful to have them show the last time the process was 
// used for something nontrivial.  The status report of "ok" indicates that
// the check was successful.
    Date saveLUT = lastUseTime;
// END KM

    try {
      exec("2+2",p);
      return( s.toString().trim().equals("4") );
    } catch (Exception e) {
      return(false);
// KM
    } finally {
      lastUseTime = saveLUT;
// END KM
    }
  }

  public static String quote(String s) {
    if (s == null) {
      return("NULL");
    }
    int len = s.length();
    char c[] = new char[len];
    s.getChars(0,len,c,0);
    char q[] = new char[2*len + 2];
    q[0] = '\"';
    int j = 1;
    for (int i = 0; i < len; i++) {
      switch(c[i]) {
      case '\\':
        q[j] = '\\'; q[j+1] = '\\'; j += 2; break; 
      case '\"':
        q[j] = '\\'; q[j+1] = '\"'; j += 2; break;
      case '\n':
        q[j] = '\\'; q[j+1] = 'n';  j += 2; break;
      case '\t':
        q[j] = '\\'; q[j+1] = 't';  j += 2; break;
      case '\r':
        q[j] = '\\'; q[j+1] = 'r';  j += 2; break;
      default:
        q[j] = c[i];
        j++;
      }
    }
    q[j] = '\"';
    return(new String(q,0,j+1));
  }
}

// Local Variables:
// c-basic-offset: 2
// compile-command: "javac -d /aim/WEB-INF/classes Maple.java"
// End:
