/*
 * @(#)MaplePool.java  May 2001
 *
 * This class maintains a pool of Maple processes.  The method 
 * getMaple() returns a Maple object representing a maple process that
 * is currently unused.  A new process will be started if there are none
 * available in the pool.  When you have finished using it, you should
 * pass it to the returnMaple() method to return it to the pool.
 *
 * TODO: better handling of broken or nonresponding processes
 */

import java.io.*;
import java.util.*;
import java.text.DateFormat;

public class MaplePool {
  public String program = null;     // absolute path to the Maple program
  public String initOptions = null; // initialization options for Maple
  public String logFile = null;     // log file
  public int maxProcesses = 8;      // max allowed number of processes

// KM
  public String killProgram = null; // absolute path to the kill program
  public String killProgramArg = null; // argument to the kill program that precedes the PID
// END KM

  public Hashtable maplesByID;
  public Hashtable maplesByStudent;
  public Hashtable lastAccessByStudent;

  public int nextID;

  public MaplePool() {
    nextID = 0;
    maplesByID = new Hashtable();
    maplesByStudent = new Hashtable();
    lastAccessByStudent = new Hashtable();
  }

  public MaplePool(String program) {
    this();
    if (program  != null) {
      this.program = program;
    }
  }

  public MaplePool(String program,String initoptions) {
    this();
    if (program != null ) {
      this.program = program;
    }
    if (initoptions != null) {
      this.initOptions = initoptions;
    }
  }

// KM
// DIST  private Maple startMaple() 
  private synchronized Maple startMaple() 
// END KM
  // Start a new Maple process, mark it as busy, add it to the pool,
  // and return it.  This method does not check the limit on process
  // numbers.
    throws IOException {
// KM
// DIST    synchronized (this) {
// DIST      Maple maple = new Maple(program,initOptions);
      Maple maple = new Maple(program,initOptions,killProgram,killProgramArg);
// END KM

      maple.id = nextID++;
      maple.busy = true;
      maple.user = null;
      maple.load = 0;
      maplesByID.put(new Integer(maple.id),maple);
      writeLog("Starting new Maple process: ID = " + maple.id + "\n");
      return maple;
// KM
// DIST    }
// END KM

  }

// KM
// DIST  public Maple getMaple(String student) 
  public synchronized Maple getMaple(String student) 
// END KM
  // Get a Maple process for use by the specified student.  The pool will
  // remember which process the student has used, and try to assign the
  // same one in future.  This maximizes the benefit from Maple's
  // caching of question versions.
    throws IOException {
    Maple maple,oldmaple,bestmaple;
    Enumeration ids;
    Integer id;
    int minload;

// KM
// DIST    synchronized(this) {
// END KM
      // remember the time.  
      // (Any process assigned to a student will be unassigned if it
      // has not been used for a certain length of time.)
      lastAccessByStudent.put(student,new Date());

      // get the process assigned to this student (if any).
      bestmaple = (Maple)maplesByStudent.get(student);
      
      if (bestmaple == null) {
        // no process assigned to this student yet

        if (maplesByID.size() < maxProcesses) {
          // under the process limit, so start a new one for this student
          bestmaple = startMaple();
          bestmaple.load = 1;
          maplesByStudent.put(student,bestmaple);
          bestmaple.busy = true;
          bestmaple.user = student;
          return(bestmaple);
        } else {
          // no more processes allowed, so find one that this student
          // can share, with the minimum possible load
          ids = maplesByID.keys();
          minload = Integer.MAX_VALUE;
          while (ids.hasMoreElements()) {
            id = (Integer)ids.nextElement();
            maple = (Maple)maplesByID.get(id);
            if ((! maple.busy) && (maple.load < minload)) {
              minload = Math.min(minload,maple.load);
              bestmaple = maple;
            }
          }

          if (bestmaple == null) {
            writeLog("Servlet busy - request refused\n");
            return(null);
          } else {
            bestmaple.busy = true;
            bestmaple.load++;
            maplesByStudent.put(student,bestmaple);
            bestmaple.user = student;
            return(bestmaple);
          }
        }
      } else {
        // there is a process assigned to this student

        if (bestmaple.busy) {
          // the assigned process is busy 

          if (! student.equals("(admin)") &&
              student.equals(bestmaple.user) &&
              ((new Date()).getTime() - 
               bestmaple.lastUseTime.getTime()) < 30000) {
            // The process is still working on the last request from
            // this student, which was submitted less than 30 seconds
            // ago; so we reject the request.
            return(null);
          }

          // find another process that this student can use
          ids = maplesByID.keys();
          oldmaple = bestmaple;
          bestmaple = null;
          minload = Integer.MAX_VALUE;
          while (ids.hasMoreElements()) {
            id = (Integer)ids.nextElement();
            maple = (Maple) maplesByID.get(id);
            if ((! maple.busy) && (maple.load < minload)) {
              minload = Math.min(minload,maple.load);
              bestmaple = maple;
            }
          }

          if (bestmaple == null) {
            // 
            if (maplesByID.size() >= maxProcesses) {
              writeLog("Servlet busy - request refused\n");
              return(null);
            } else {
              bestmaple = startMaple();
            }
          }

          bestmaple.busy = true;
          bestmaple.user = student;
          oldmaple.load--;
          bestmaple.load++;
          maplesByStudent.put(student,bestmaple);
          return(bestmaple);
        } else {
          // assigned process is not busy
          bestmaple.busy = true;
          bestmaple.user = student;
          return(bestmaple);
        }
      }
// KM
// DIST    }
// END KM
  }

// KM
  // removes the maple object from the pool, 
  // but doesn't kill the process if its running
// this is only called by other higher level routines,
// so I am changing it from public to private
// dist  public void removeMaple(Maple maple) {
  private synchronized void removeMaple(Maple maple) {
// END KM
    Enumeration students;
    String student;
// KM    
// DIST    synchronized(this) {
// END KM
      maplesByID.remove(new Integer(maple.id));
      students = maplesByStudent.keys();
      while (students.hasMoreElements()) {
        student = (String)students.nextElement();
        if (maplesByStudent.get(student) == maple) {
          maplesByStudent.remove(student);
        }
      }
// KM
// DIST    }
// END KM
  }

  // returns a Maple object to the pool when its done executing a command 
// KM
// DIST  public void returnMaple(Maple maple) 
  public synchronized void returnMaple(Maple maple) 
// END KM
    throws IOException {
// KM
// dist    Enumeration students;
// dist    String student;
// dist    synchronized(this) {
// END KM
    if (maple.condemned) {
      removeMaple(maple);
      writeLog("Closing condemned Maple process: ID = " + maple.id +"\n");
      try {maple.close();}
      catch(Exception e) {};
    } else {
      maple.busy = false;
      maple.user = null;
    }
  }

  // This removes the Maple object from the pool and unconditionally kills its process
// KM
// DIST  public void discardMaple(Maple maple) {
// END KM
  public void discardMaple(Maple maple) {

// KM
// dist    Enumeration students;
// dist    String student;
// END KM
    
      removeMaple(maple);
      try {
// KM
// dist      writeLog("Closing condemned Maple process: ID = " + maple.id +"\n");
        writeLog("Destroying condemned Maple process: ID = " + maple.id +"\n");
// END KM
        maple.destroy();
      }
      catch(Exception e) {};
// KM
      maple.discarded = true;
// END KM
  }

  // Condemn the maple process and if its not busy, close it.
// KM
// DIST  public void condemn(Maple maple) {
  public synchronized void condemn(Maple maple) {
// DIST    synchronized(this) {
// END KM
      maple.condemned = true;
      if ( ! maple.busy) {
        removeMaple(maple);
        try {
          maple.close();
        } catch(Exception e) {};
      }
// KM
// DIST    }
// END KM
  }

  // close all the processes in the pool
// KM
// DIST  public void drain() 
  public synchronized void drain() 
// END KM
  throws IOException {
    Maple maple;
    Integer id;
    Enumeration ids;

    writeLog("Draining Maple process pool");
// KM
// DIST    synchronized(this) {
// END KM
    ids = maplesByID.keys();
    while (ids.hasMoreElements()) {
      id = (Integer)ids.nextElement();
      maple = (Maple)maplesByID.get(id);
      maple.close();
    }
    maplesByID.clear();
    maplesByStudent.clear();
// KM
// DIST    }
// END KM

  }

  // destroy all the processes in the pool
// KM
// DIST  public void scorch() 
  public synchronized void scorch() 
// END KM
  throws IOException {
    Maple maple;
    Integer id;
    Enumeration ids;

    writeLog("Scorching Maple process pool");
// KM
// DIST    synchronized(this) {
// END KM
      ids = maplesByID.keys();
      while (ids.hasMoreElements()) {
        id = (Integer)ids.nextElement();
        maple = (Maple)maplesByID.get(id);
// KM
        maple.discarded = true;
// END KM
        maple.destroy();
      }
      maplesByID.clear();
      maplesByStudent.clear();
// KM
// DIST    }
// END KM
  }

  // close all processes that have not been used for age0 milliseconds,
  // or that were created more than age1 milliseconds ago.  Also
  // unassign processes from students who have not accessed the
  // system for age0 milliseconds.
// KM
// DIST  public void clean(long age0,long age1) 
  public synchronized void clean(long age0,long age1) 
// END KM
    throws IOException {
    long cutoff0 = (new Date()).getTime() - age0;
    long cutoff1 = (new Date()).getTime() - age1;
    Maple maple;
    Enumeration ids,students;
    Integer id;
    String student;
    StringBuffer msg = new StringBuffer();

    msg.append("Cleaning Maple process pool\n");
// KM
// DIST    synchronized(this) {
// END KM
    ids = maplesByID.keys();
      
    while(ids.hasMoreElements()) {
      id = (Integer)ids.nextElement();
      maple = (Maple) maplesByID.get(id);
      if (maple.lastUseTime.getTime() < cutoff0 ||
          maple.creationTime.getTime() < cutoff1) {
        msg.append("Condemning old process " + maple.id + "\n");
        condemn(maple);
      }
    }

    students = maplesByStudent.keys();
    while (students.hasMoreElements()) {
      student = (String) students.nextElement();
      if (((Date)lastAccessByStudent.get(student)).getTime() < cutoff0) {
        lastAccessByStudent.remove(student);
        maple = (Maple) maplesByStudent.get(student);
        if (maple != null) {
          maplesByStudent.remove(student);
          maple.load--;
        }
      }
    }
    writeLog(msg.toString());
// KM
// DIST    }
// END KM
  }

  private void writeLog(String s) 
    throws IOException {
    if (logFile != null) {
      PrintWriter logWriter = new PrintWriter(new FileWriter(logFile,true));
      DateFormat df = DateFormat.getDateTimeInstance();
      logWriter.println(df.format(new Date()));
      logWriter.println(s);
      logWriter.close();
    }
  }

// KM
// DIST  public String statusReport() {
  public synchronized String statusReport() {
// END KM
    StringBuffer report = new StringBuffer();
    Enumeration ids,students;
    Integer id;
    String student;
    Maple maple = null;
    DateFormat tf =  DateFormat.getTimeInstance();
    String creationtime,lastusetime,user,status;
    Hashtable index;
// KM
// DIST    synchronized (this) {
// END KM
      ids = maplesByID.keys();
      if (ids.hasMoreElements()) {
        report.append(
        "<table border width='100%'>\n" +
        " <tr>\n" +
        "  <th>ID</th>\n" +

// KM
        "  <th>PID</th>\n" +
// END KM

        "  <th>Created</th>\n" +
        "  <th>Last Used</th>\n" +
        "  <th>User</th>\n" + 
        "  <th>Load</th>\n" +
        "  <th>Status</th>\n" +
        " </tr>\n");
 
        index = new Hashtable();

        while (ids.hasMoreElements()) {
          id = (Integer) ids.nextElement();
          index.put(id,new Vector());
          maple = (Maple)maplesByID.get(id);
          creationtime = tf.format(maple.creationTime);
          lastusetime = tf.format(maple.lastUseTime);
   
          if (maple.busy) {
            user = maple.user;
            status = "busy";
          } else {
            user = "(free)";
            if (maple.condemned) {
              status = "condemned";
            } else {
              if (maple.check()) {
                status = "ok";
              } else {
                status = "not ok - closing";
                condemn(maple);
              }
            }
          }
         
          report.append(" <tr>\n" + 
          "  <td>" + maple.id + "</td>\n" +
// KM
          "  <td>" + (maple.pid == -1 ? "N/A" : Integer.toString(maple.pid)) + "</td>\n" +
// END KM
          "  <td>" + creationtime + "</td>\n" +
          "  <td>" + lastusetime + "</td>\n" +
          "  <td>" + user + "</td>\n" +
          "  <td>" + maple.load + "</td>\n" +
          "  <td>" + status + "</td>\n" +
         " </tr>\n");
 
        }
        report.append("</table>\n");

        students = maplesByStudent.keys();
        while (students.hasMoreElements()) {
          student = (String) students.nextElement();
          id = new Integer(((Maple) (maplesByStudent.get(student))).id);
          ((Vector) index.get(id)).addElement(student);
        }

        report.append("<h2>Process assignments:</h2>\n");
        report.append("<table border width='100%'>\n" +
                      "<tr>\n" +
                      "<th width='10%'>Process</th>\n" +
                      "<th width='90%'>Students</th>\n" );
        ids = maplesByID.keys();
        while (ids.hasMoreElements()) {
          id = (Integer) ids.nextElement();
          report.append(
          "<tr><td width='10%'>" +
          id.intValue() +
          "</td>\n<td width='90%'>");
          students = ((Vector) index.get(id)).elements();
          while (students.hasMoreElements()) {
            student = (String) students.nextElement();
            report.append(student + "&nbsp;");
          }
          report.append("</td></tr>\n");
        }
        report.append("</table>\n");
      } else {
        report.append("There are no active Maple processes.<br>\n");
      }
      
      return (report.toString());
    }
// KM
// DIST  }
// END KM

}

// Local Variables:
// c-basic-offset: 2
// compile-command: "javac -d /aim/WEB-INF/classes MaplePool.java"
// End:
